from collections import namedtuple

from .advanced_query import AQ
from .document import Document

__all__ = ("AQ", "Document")
__version__ = "0.0.3"
VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(major=0, minor=0, micro=3, releaselevel="final", serial=0)



import asyncio
import os

from motor.motor_asyncio import AsyncIOMotorClient

from alaric import *
from alaric.comparison import *
from alaric.logical import *
from alaric.meta import *


async def main():
    client = AsyncIOMotorClient(os.environ["MONGO"])
    db = client["COMPX323"]
    document: Document = Document(db, "movies")

    # 1) {'director': {'$eq': 'Steven Spielberg'}}
    r_1 = await document.find_many(AQ(EQ("director", "Steven Spielberg")))
    print(r_1)

    # 2) {'actor': {'$eq': {'name': 'Mark Hamill'}}}
    r_2 = await document.find_many(AQ(EQ("actor", {"name": "Mark Hamill"})))
    print(r_2)

    # 3) {'$and': [{'director': {'$eq': 'Quentin Tarantino'}}, {'actor': {'$eq': {'name': 'Quentin Tarantino'}}}]}
    r_3 = await document.find_many(
        AQ(
            AND(
                EQ("director", "Quentin Tarantino"),
                EQ("actor", {"name": "Quentin Tarantino"}),
            ),
        )
    )
    print(r_3)

    # 4) {"$query": {}, "$orderby": {"year": 1}}
    r_4 = await document.find_many({"$query": {}, "$orderby": {"year": 1}})
    print(r_4)

    # 5) {'genre': {'$in': ['action', 'fantasy']}}
    r_5 = await document.find_many(AQ(IN("genre", ["action", "fantasy"])))
    print(r_5)

    # 6) {'year': {'$gt': 2005}}
    r_6 = await document.find_many(AQ(GT("year", 2005)))
    print(r_6)

    # 7) {'$and': [{'director': {'$eq': 'Quentin Tarantino'}}, {'$or': [{'actor': {'$eq': {'name': 'Christoph Waltz'}}}, {'actor': {'$eq': {'name': 'Leonardo DiCaprio'}}}]}]}
    r_7 = await document.find_many(
        AQ(
            AND(
                EQ("director", "Quentin Tarantino"),
                OR(
                    EQ("actor", {"name": "Christoph Waltz"}),
                    EQ("actor", {"name": "Leonardo DiCaprio"}),
                ),
            )
        )
    )
    print(r_7)

    # 8) {'director': {'$eq': 'Quentin Tarantino'}}
    r_8 = await document.count(EQ("director", "Quentin Tarantino"))
    print(r_8)

    # 9) {'year': {'$lt': 2000}}
    r_9 = await document.count(LT("year", 2000))
    print(r_9)

    # Un-needed after initial run
    # await document.bulk_insert(
    #     [
    #         {
    #             "title": "Star Wars",
    #             "year": 1977,
    #             "director": "George Lucas",
    #             "genre": "fantasy",
    #             "actor": [
    #                 {"name": "Mark Hamill"},
    #                 {"name": "Carrie Fisher"},
    #                 {"name": "Harrison Ford"},
    #             ],
    #         },
    #         {
    #             "title": "Blade Runner",
    #             "year": 1982,
    #             "director": "Ridley Scott",
    #             "genre": "scifi",
    #             "actor": [{"name": "Harrison Ford"}],
    #         },
    #         {
    #             "title": "The Empire Strikes Back",
    #             "year": 1980,
    #             "director": "Irvin Kershner",
    #             "genre": "fantasy",
    #             "actor": [
    #                 {"name": "Mark Hamill"},
    #                 {"name": "Carrie Fisher"},
    #                 {"name": "Harrison Ford"},
    #             ],
    #         },
    #         {
    #             "title": "Alien",
    #             "year": 1979,
    #             "director": "Ridley Scott",
    #             "genre": "scifi",
    #             "actor": [{"name": "Sigourney Weaver"}],
    #         },
    #         {
    #             "title": "Aliens",
    #             "year": 1986,
    #             "director": "James Cameron",
    #             "genre": "scifi",
    #             "actor": [{"name": "Sigourney Weaver"}],
    #         },
    #         {
    #             "title": "Toy Story",
    #             "year": 1995,
    #             "director": "John Lasseter",
    #             "genre": "fantasy",
    #             "actor": [{"name": "Tom Hanks"}, {"name": "Tim Allen"}],
    #         },
    #         {
    #             "title": "Schindler's List",
    #             "year": 1993,
    #             "director": "Steven Spielberg",
    #             "genre": "drama",
    #             "actor": [{"name": "Liam Neeson"}],
    #         },
    #         {
    #             "title": "E.T.",
    #             "year": 1982,
    #             "director": "Steven Spielberg",
    #             "genre": "fantasy",
    #             "actor": [{"name": "Henry Thomas"}],
    #         },
    #         {
    #             "title": "Raiders of the Lost Ark",
    #             "year": 1981,
    #             "director": "Steven Spielberg",
    #             "genre": "action",
    #             "actor": [{"name": "Harrison Ford"}],
    #         },
    #         {
    #             "title": "Minority Report",
    #             "year": 2002,
    #             "director": "Steven Spielberg",
    #             "genre": "scifi",
    #             "actor": [{"name": "Tom Cruise"}],
    #         },
    #         {
    #             "title": "The Departed",
    #             "year": 2006,
    #             "director": "Martin Scorsese",
    #             "genre": "thriller",
    #             "actor": [
    #                 {"name": "Leonardo Dicaprio"},
    #                 {"name": "Jack Nicholson"},
    #                 {"name": "Matt Damon"},
    #                 {"name": "Mark Wahlberg"},
    #             ],
    #         },
    #         {
    #             "title": "The Prestige",
    #             "year": 2006,
    #             "director": "Christopher Nolan",
    #             "genre": "thriller",
    #             "actor": [
    #                 {"name": "Hugh Jackman"},
    #                 {"name": "Christian Bale"},
    #                 {"name": "Scarlett Johansson"},
    #             ],
    #         },
    #         {
    #             "title": "Reservoir Dogs",
    #             "year": 1992,
    #             "director": "Quentin Tarantino",
    #             "genre": "thriller",
    #             "actor": [
    #                 {"name": "Harvey Keitel"},
    #                 {"name": "Tim Roth"},
    #                 {"name": "Quentin Tarantino"},
    #             ],
    #         },
    #         {
    #             "title": "Inglourious Basterds",
    #             "year": 2009,
    #             "director": "Quentin Tarantino",
    #             "genre": "war",
    #             "actor": [
    #                 {"name": "Brad Pitt"},
    #                 {"name": "Christoph Waltz"},
    #                 {"name": "Melanie Laurent"},
    #             ],
    #         },
    #         {
    #             "title": "Letters from Iwo Jima",
    #             "year": 2006,
    #             "director": "Clint Eastwood",
    #             "genre": "war",
    #             "actor": [{"name": "Ken Watanabe"}],
    #         },
    #         {
    #             "title": "Slumdog Millionaire",
    #             "year": 2008,
    #             "director": "Danny Boyle",
    #             "genre": "drama",
    #             "actor": [{"name": "Dev Patel"}],
    #         },
    #         {
    #             "title": "Django Unchained",
    #             "year": 2012,
    #             "director": "Quentin Tarantino",
    #             "genre": "action",
    #             "actor": [
    #                 {"name": "Jamie Foxx"},
    #                 {"name": "Christoph Waltz"},
    #                 {"name": "Leonardo Dicaprio"},
    #             ],
    #         },
    #         {
    #             "title": "The Dark Knight",
    #             "year": 2008,
    #             "director": "Christopher Nolan",
    #             "genre": "thriller",
    #             "actor": [
    #                 {"name": "Christian Bale"},
    #                 {"name": "Heath Ledger"},
    #                 {"name": "Michael Caine"},
    #             ],
    #         },
    #         {
    #             "title": "Memento",
    #             "year": 2005,
    #             "director": "Christopher Nolan",
    #             "genre": "thriller",
    #             "actor": [{"name": "Guy Pearce"}],
    #         },
    #         {
    #             "title": "127 Hours",
    #             "year": 2010,
    #             "director": "Danny Boyle",
    #             "genre": "drama",
    #             "actor": [{"name": "James Franco"}],
    #         },
    #         {
    #             "title": "Birdman",
    #             "year": 2014,
    #             "director": "Alejandro Gonzalez Innaritu",
    #             "genre": "comedy",
    #             "actor": [{"name": "Michael Keaton"}, {"name": "Edward Norton"}],
    #         },
    #         {
    #             "title": "The Shining",
    #             "year": 1980,
    #             "director": "Stanley Kubrick",
    #             "genre": "horror",
    #             "actor": [{"name": "Jack Nicholson"}],
    #         },
    #         {
    #             "title": "Kill Bill Volume 1",
    #             "year": 2003,
    #             "director": "Quentin Tarantino",
    #             "genre": "action",
    #             "actor": [{"name": "Uma Thurman"}],
    #         },
    #         {
    #             "title": "La La Land",
    #             "year": 2016,
    #             "director": "Damien Chazelle",
    #             "genre": "romance",
    #             "actor": [{"name": "Emma Stone"}, {"name": "Ryan Gosling"}],
    #         },
    #     ]
    # )


if __name__ == "__main__":
    asyncio.run(main())



import functools
from typing import List, Dict, Optional, Union, Any, TypeVar, Type

from pymongo.results import DeleteResult
from motor.motor_asyncio import AsyncIOMotorDatabase, AsyncIOMotorCollection

from alaric.abc import BuildAble

T = TypeVar("T")


class Document:
    _version = 10

    def __init__(
        self,
        database: AsyncIOMotorDatabase,
        document_name: str,
        converter: Optional[Type[T]] = None,
    ):
        """
        Parameters
        ----------
        database: AsyncIOMotorDatabase
            The database we are connected to
        document_name: str
            What this _document should be called
        converter: Optional[Type[T]]
            An optional class to try
            to convert all data-types which
            return either Dict or List into
        """
        self._document_name: str = document_name
        self._database: AsyncIOMotorDatabase = database
        self._document: AsyncIOMotorCollection = database[document_name]

        self.converter: Type[T] = converter

    def __repr__(self):
        return f"<Document(document_name={self.collection_name})>"

    async def find(
        self, filter_dict: Union[Dict[str, Any], BuildAble]
    ) -> Optional[Union[Dict[str, Any], Type[T]]]:
        """Find and return one item.

        Parameters
        ----------
        filter_dict: Union[Dict, BuildAble]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        Optional[Union[Dict[str, Any], Type[T]]]
            The result of the query
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        data = await self._document.find_one(filter_dict)
        return await self.attempt_convert(data)

    async def find_many(
        self, filter_dict: Union[Dict[str, Any], BuildAble]
    ) -> List[Union[Dict[str, Any], Type[T]]]:
        """
        Find and return all items
        matching the given filter.

        Parameters
        ----------
        filter_dict: Dict[str, Any]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        List[Union[Dict[str, Any], Type[T]]]
            The result of the query
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        data = await self._document.find(filter_dict).to_list(None)
        return await self.attempt_convert(data)

    async def delete(
        self, filter_dict: Union[Dict, BuildAble]
    ) -> Optional[DeleteResult]:
        """
        Delete an item from the Document
        if an item with the provided filter exists.

        Parameters
        ----------
        filter_dict: Union[Dict, BuildAble]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        Optional[DeleteResult]
            The result of deletion if it occurred.
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        result: DeleteResult = await self._document.delete_many(filter_dict)
        result: Optional[DeleteResult] = result if result.deleted_count != 0 else None
        return result

    async def get_all(
        self,
        filter_dict: Optional[Union[Dict[str, Any], BuildAble]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> List[Optional[Union[Dict[str, Any], Type[T]]]]:
        """
        Fetches and returns all items
        which match the given filter.

        Example usages

        .. code-block:: python
            results = await document.get_all()

            ...

            # Get all documents with a given field
            results = await document.get_all(AQ(EXISTS("field")))

        Parameters
        ----------
        filter_dict: Optional[Dict[str, Any]]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        List[Optional[Union[Dict[str, Any], Type[T]]]]
            The items matching the filter
        """
        filter_dict = filter_dict or {}
        filter_dict = self.__ensure_built(filter_dict)

        data = await self._document.find(filter_dict, *args, **kwargs).to_list(None)
        return await self.attempt_convert(data)

    async def insert(self, data: Dict[str, Any]) -> None:
        """Insert the provided data into the document.

        Parameters
        ----------
        data: Dict[str, Any]
            The data to insert
        """
        self.__ensure_dict(data)
        await self._document.insert_one(data)

    async def update(
        self,
        filter_dict: Union[Dict[str, Any], BuildAble],
        update_data: Dict[str, Any],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Performs an UPDATE operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], BuildAble]
            The data to filter on
        update_data: Dict[str, Any]
            The data to upsert
        option: str
            The optional option to pass to mongo,
            default is set
            https://www.mongodb.com/docs/manual/reference/operator/update/
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        self.__ensure_dict(update_data)

        await self._document.update_one(
            filter_dict, {f"${option}": update_data}, *args, **kwargs
        )

    async def upsert(
        self,
        filter_dict: Union[Dict[str, Any], BuildAble],
        update_data: Dict[str, Any],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Performs an UPSERT operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], BuildAble]
            The data to filter on
        update_data: Dict[str, Any]
            The data to upsert
        option: str
            Update operator.
            https://www.mongodb.com/docs/manual/reference/operator/update/
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        self.__ensure_dict(update_data)
        await self.update(
            filter_dict, update_data, option, upsert=True, *args, **kwargs
        )

    async def unset(
        self, filter_dict: Union[Dict[str, Any], BuildAble], field: Any
    ) -> None:
        """Delete a given field on a collection

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], BuildAble]
            The fields to match on (Think _id)
        field: Any
            The field to remove
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        await self._document.update_one(filter_dict, {"$unset": {field: True}})

    async def increment(
        self,
        filter_dict: Union[Dict[str, Any], BuildAble],
        field: str,
        amount: Union[int, float],
    ) -> None:
        """Increment the provided field.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], BuildAble]
            The 'thing' we want to increment
        field: str
            The key for the field to increment
        amount: Union[int, float]
            How much to increment (or decrement) by
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        await self._document.update_one(filter_dict, {"$inc": {field: amount}})

    async def change_field_to(
        self,
        filter_dict: Union[Dict[str, Any], BuildAble],
        field: str,
        new_value: Any,
    ) -> None:
        """Modify a single field and change the value.

        Parameters
        ----------
        filter_dict: Union[Dict[Any, Any], BuildAble]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        field: str
            The key for the field to increment
        new_value: Any
            What the field should get changed to
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        await self._document.update_one(filter_dict, {"$set": {field: new_value}})

    async def count(self, filter_dict: Union[Dict[Any, Any], BuildAble]) -> int:
        """Return a count of how many items match the filter.

        Parameters
        ----------
        filter_dict:  Union[Dict[Any, Any], BuildAble]
            The count filer.

        Returns
        -------
        int
            How many items matched the filter.
        """
        filter_dict = self.__ensure_built(filter_dict)
        self.__ensure_dict(filter_dict)
        return await self._document.count_documents(filter_dict)

    async def bulk_insert(self, data: List[Dict]) -> None:
        """
        Given a List of Dictionaries, bulk insert all
        the given dictionaries in a single call.

        Parameters
        ----------
        data: List[Dict]
            The data to bulk insert
        """
        self.__ensure_list_of_dicts(data)
        await self._document.insert_many(data)

    # <-- Private methods -->
    @staticmethod
    def __ensure_list_of_dicts(data: List[Dict]):
        assert isinstance(data, list)
        assert all(isinstance(entry, dict) for entry in data)

    @staticmethod
    def __ensure_dict(data: Dict[str, Any]) -> None:
        assert isinstance(data, dict)

    @staticmethod
    def __ensure_built(data: Union[Dict, BuildAble]) -> Dict:
        if isinstance(data, BuildAble):
            return data.build()

        return data

    async def attempt_convert(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[List[Union[Dict[str, Any], Type[T]]], Union[Dict[str, Any], Type[T]]]:
        # This exists purely so we don't lose intelli-sense
        # as decorators remove intelli-sense
        if not data or not self.converter:
            return data

        if not isinstance(data, list):
            return self.converter(**data)

        new_data = []
        for d in data:
            new_data.append(self.converter(**d))

        return new_data

    # <-- Some basic internals -->
    @property
    def collection_name(self) -> str:
        """The connected collections name."""
        return self._document_name

    @property
    def document_name(self) -> str:
        return self._document_name

    @property
    def raw_database(self) -> AsyncIOMotorDatabase:
        """Access to the database instance."""
        return self._database

    @property
    def raw_collection(self) -> AsyncIOMotorCollection:
        """The connection collection instance."""
        return self._document



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from alaric.comparison import IN, GT, LT
from alaric.comparison.comparison_exists import EXISTS

if TYPE_CHECKING:
    from alaric.abc import ComparisonT


class NEGATE:
    """
    Negate a given option, I.e. Do the opposite.

    Supported operands:

    * EXISTS
    * IN
    """

    def __init__(self, comparison: ComparisonT):
        self.comparison: ComparisonT = comparison

    def __repr__(self):
        return f"NEGATE({self.comparison})"

    def build(self) -> Dict:
        if isinstance(self.comparison, EXISTS):
            self.comparison._val = False
            return self.comparison.build()

        elif isinstance(self.comparison, IN):
            self.comparison._operator = "$nin"
            return self.comparison.build()

        elif isinstance(self.comparison, GT):
            raise RuntimeError("Cannot negate GT, use LT instead.")

        elif isinstance(self.comparison, LT):
            raise RuntimeError("Cannot negate LT, use GT instead.")

        raise RuntimeError("Invalid wrapped comparison.")



from .comparison_eq import EQ
from .comparison_in import IN
from .comparison_exists import EXISTS
from .comparison_less_than import LT
from .comparison_greater_than import GT

__all__ = ("EQ", "IN", "EXISTS", "GT", "LT")
